/***********************************/
/* Attila Gyenesei & Jukka Teuhola */
/*                                 */     
/*          PIE algorithm          */
/***********************************/

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <fstream.h>
#include <string.h>
#include <math.h>

#define MaxHashSize  4096
#define Mask         4095
#define testing false

typedef struct node node;
struct node{
        int freq; //frequency of the node
        float err; //error of the node
        float ce; //children errors of the node
        bool ready;
        struct node **children; //next children
    }*root;

struct OfNode {
    int itemNo;
    int f;
    struct OfNode *of;
};

struct OfNode *Overflow[MaxHashSize]; //for hash table
int ItemNo[MaxHashSize], freq[MaxHashSize]; //for hash table
FILE *Fin, *Fout; //for input-output processes
int *Item; //items vector
int Itemsno; //number of items
int *Map, *Remap; //mapping vector for items
int *Trans; //vector list of current transaction
int Minsup; //minimum support threshold
float Rminsup; //relative minimum support threshold
int Tno; //number of total transactions
int Fno; //number of frequent items
int Nodesno; //number of nodes in the tree
int Freqsetsno; //number of frequent itemsets
int Readynodesno; //number of ready nodes
char *Inputfilename, *Foutfilename;
float Oldalfa, Alfa; // average probability of frequent items
int Iteration; //number of iteration;
bool Outputfile; //result to the output file
int Level[Mask]; //for result
int Emptytrans; //number of empty transactions
int ItemCount; //number of (real) items in the dataset
int MaxItemsNo; //maximal number of item in the dataset


//*************************************************************************
//* Initialize hash table                                                 *
//*************************************************************************
void init() {
    int i;

    Tno = 0;
    ItemCount = 0;
    Emptytrans = 0;
    MaxItemsNo = 0;
    for (i=0; i<MaxHashSize; i++) {
        ItemNo[i] = -1;
        freq[i] = 0;
    }
}
/*-----------------------------------------------------------*/


//*************************************************************************
//* Add item in the hash structure                                        *
//*************************************************************************
void addItem(int item) {
    struct OfNode *ptr, *next;
    int addr;

    addr = item & Mask;
    if (ItemNo[addr]==-1) {
        /* New item */
        ItemNo[addr] = item;
        freq[addr] = 1;
        Overflow[addr] = NULL;
        ItemCount++;
        if (item > MaxItemsNo) MaxItemsNo = item;
    }
    else if (ItemNo[addr]==item) {
            /* Hit, no overflow scan */
            freq[addr]++;
        }
        else {
            next = Overflow[addr];
            if (next == NULL) {
                /* New overflow node to the front */
                ptr = (struct OfNode *) malloc(sizeof(struct OfNode));
                Overflow[addr] = ptr;
                ptr->itemNo = item;
                ptr->f = 1;
                ptr->of = NULL;
                ItemCount++;
                if (item > MaxItemsNo) MaxItemsNo = item;
            } // end of next==NULL
            else {
            // Non-empty overflow list
            while ((next->itemNo) != item) {
                ptr = next;
                next = ptr->of;
                if (next==NULL) {
                    // New node to the end
                    next = (struct OfNode *) malloc(sizeof(struct OfNode));
                    ptr->of = next;
                    next->itemNo = item;
                    next->f = 0;
                    next->of = NULL;
                    ItemCount++;
                    if (item > MaxItemsNo) MaxItemsNo = item;
                } // end of next==NULL
            } // end of item not found
            (next->f)++;
        } // end of non-empty overflow list
    } // end of handling overflow list
} // end of addItem
/*-----------------------------------------------------------*/


//*************************************************************************
//* Reads the transactions and stores the item frequencies                *
//* into a 'hash' table.                                                  *
//*************************************************************************
void initialReadTrans() {
    int c;
    int count, num;
    int fileready, lineready;

    fileready = 0;
    do {
        lineready = 0;
        count = 0;
        c = fgetc(Fin);
        do {
            // Skip white spaces
            while ((c<'0') || (c>'9')) {
                if (c=='\n') {
                    lineready = 1;
                    break;
                }
                if (c==EOF) {
                    lineready = 1;
                    fileready = 1;
                    break;
                }
                c = fgetc(Fin);
            } // end of not number
            if (!lineready) {
                // Read item number
                num = c-('0');
                c = fgetc(Fin);
                while ((c>='0') && (c<='9')) {
                    num = 10*num + c-'0';
                    c = fgetc(Fin);
                }
                addItem(num);
                count++;
                if (c=='\n') lineready = 1;
                else if (c==EOF) {
                    lineready = 1;
                    fileready = 1;
                }
            } // end of not lineready
        } while (!lineready);
        if (count==0) {
            if (!fileready) {
                Emptytrans++;
                Tno++;
            }
        }
        else{Tno++;}
    } while (!fileready);
    return;
}  // end of ititialReadTrans
/*-----------------------------------------------------------*/


//*************************************************************************
//* Returns the number of frequent items, and those items in a global     *
//* mapping table (Map).                                                  *
//*************************************************************************
void findFreqItems() {
    int i;
    struct OfNode *ptr;

    Item = (int *) calloc(MaxItemsNo+1, sizeof(int));
    for (i=0; i<MaxHashSize; i++) {
        if (ItemNo[i] != -1) {
            if (freq[i] >= Minsup) {
                Item[ItemNo[i]] = freq[i];
            }
            ptr = Overflow[i];
            while (ptr != NULL) {
                if ((ptr->f) >= Minsup) {
                    Item[ptr->itemNo] = (ptr->f);
                }
                ptr = ptr->of;
            }
        }
    }
    return;
} // end of findFreqItems
/*-----------------------------------------------------------*/


//*************************************************************************
//* Read the input datafile (using hash table) and set variables          *
//*************************************************************************
void read_inputdata(){
    int i;

    if ( ( Fin = fopen(Inputfilename,"rt") ) == NULL){ //open datafile
                cout << "Datafile doesn't exist!";
                exit(2);
    }

    //input data:
    init();
    initialReadTrans();
    findFreqItems();
    fclose(Fin); //close datafile
    
    Rminsup = ((float)Minsup/(float)Tno); // relative minimum support
    Map = (int *) calloc(MaxItemsNo+1, sizeof(int));
    Remap = (int *) calloc(MaxItemsNo+1, sizeof(int));
    Itemsno = MaxItemsNo+1;
    for(i=0;i<=Itemsno-1;i++){ 
        Remap[i]=0; //inicialize the vector of Items;
        Map[i]=0; //inicialize the vector of Items;
    }
    Trans = (int *) calloc(MaxItemsNo+1, sizeof(int));

    //Mapping:    
    Alfa = 0;
    Fno = 0; 
    for(i=0;i<=Itemsno-1;i++){
        if( Item[i] < Minsup ){Map[i] = -1;}
        else{ 
            Fno++; 
            Alfa = Alfa + ((float)Item[i] / (float)Tno);
            Map[i] = Fno-1;
            Remap[Fno-1] = i;
        }
    }
    Alfa = Alfa / (float)Fno;
    Oldalfa = Alfa;
    
    if( testing ){
        cout << endl << "number of transactions = " << Tno << endl;
        cout << "number of items = " << Itemsno << endl;
        cout << "number of frequent items = " << Fno << endl;
        cout << "average prob. of frequent items = " << Alfa << endl;
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Frequency and childerror calculation                                  *
//*************************************************************************
void count_nodefreq(node *n, int act, int last){
    int i;
    bool tempce;
    
    tempce = false;
    if( !n->ready ){ n->freq++; }
    if( act != last ){
        for(i=act+1;i<=last;i++){
            if ( n->children[Trans[i]] ){
                if( !n->children[Trans[i]]->ready ){
                    count_nodefreq(n->children[Trans[i]],i,last);
                }
            }else{tempce = true;}
        }
        if( tempce ){n->ce++;}
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Inic steps for frequency and childerror calculation                   *
//*************************************************************************
void call_count_nodefreq(){
    int titemsno; //number of items in the current transaction
    char c; //for transaction reading
    int i;
    
    Fin = fopen(Inputfilename,"rt"); //open input datafile
    
    while( !feof(Fin) ){
        titemsno = 0;
        do{
            c = '0';
            i = 0;
            do{
                i *=10;
                i += int(c)-int('0');
                c = getc(Fin);
            }while(c >= '0' && c <= '9');
            if( Map[i] >= 0){
                titemsno++;
                Trans[titemsno-1] = Map[i];
            }
        }while( c!='\n' && !feof(Fin));

        //Call count_nodefreq():
        for(i=0;i<=titemsno-1;i++){
            if( !root->children[Trans[i]]->ready ){
                count_nodefreq(root->children[Trans[i]],i,titemsno-1);
            }
        }
    }
    fclose(Fin); //close datafile
}
//------------------------------------------------------------------------


//*************************************************************************
//* Set ready node variable                                               *
//*************************************************************************
void set_ready(node *n, int act)
{
    int i;
    bool r;

    if ( !n->ready ){
        r = true;
        for(i=act+1;i<=Fno-1;i++){
            if( n->children[i] ){
                set_ready(n->children[i],i);
                if( !n->children[i]->ready ){r = false;}
            }
        }
        if( (n->ce < Minsup) && ( r )){
            n->ready = true;
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Inicialize nodes' childerror and frequency variables                  *
//*************************************************************************
void inic_ce_freq(node *n, int act){
    int i;
    
    n->ce = 0;
    if( !n->ready ){
        n->freq = 0;
    }
    for(i=act+1;i<=Fno-1;i++){
        if( n->children[i] ){
            inic_ce_freq(n->children[i],i);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Frequency counting for the nodes                                      *
//*************************************************************************
void count_frequency(node *n, int act){
    int i;
    
    if( n->freq >= Minsup ){Freqsetsno++;}
    for(i=act+1;i<=Fno-1;i++){
        if( n->children[i] ){
            count_frequency(n->children[i],i);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Count ready nodes                                                     *
//*************************************************************************
void count_readynodes(node *n, int act){
    int i;
    
    if( (n != root) && (n->ready) ){Readynodesno++;}
    for(i=act+1;i<=Fno-1;i++){
        if( n->children[i] ){
            count_readynodes(n->children[i],i);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Add new node to the expanded tree                                     *
//*************************************************************************
void add_expandnode(node *n, int act, int last){
    struct node *newn;
    int i, j;
    float prob, probsum, tempperr;

    if( n->ce < Minsup ){ //just go the children
        for(i=act+1;i<=last;i++){
            if( n->children[i] ){
                if( !n->children[i]->ready ){
                    add_expandnode(n->children[i],i,last);
                }
            }
        }
    }
    else{
        //Probsum calculation:
        probsum = 1;
        for(i=act+1;i<=last;i++){
            if( !n->children[i] ){
                prob = (float)Item[Remap[i]] / (float)Tno;                
                probsum = probsum * (1.0 - prob);
            }
        }
        probsum = 1 - probsum;

        for(i=act+1;i<=last;i++){
            if( n->children[i] ){
                if( !n->children[i]->ready ){
                    add_expandnode(n->children[i],i,last);
                }
            }
            else{
                prob = (float)Item[Remap[i]] / (float)Tno;
                tempperr = (prob / probsum) * n->ce;
                if( tempperr >= Alfa * (float)Minsup ){
                    //add new node:
                    Nodesno++;
                    newn = (node *)malloc(sizeof(node));
                    n->children[i] = newn;
                    newn->freq = 0;
                    newn->ready = false;
                    newn->ce = tempperr;
                    newn->err = 0;
                    newn->children = (node**)malloc(Fno*sizeof(node)); 
                    for(j=0;j<=Fno-1;j++){newn->children[j] = NULL;}
                    add_expandnode(newn,i,last);
                }
            }
        }
    }//else
}
//------------------------------------------------------------------------


//*************************************************************************
//* Expand the tree                                                       *
//*************************************************************************
void expand_tree(){
    int i; //loop variable

    for(i=0;i<=Fno-1;i++){
        if ( !root->children[i]->ready ){
            add_expandnode(root->children[i],i,Fno-1);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Add new node the the tree (in phase 1)                                *
//*************************************************************************
void add_node(node *p, float perr, int act, int last)
{
    struct node *newn;
    int i;
    float prob;
    
    //New node:
    Nodesno++;
    newn = (node *)malloc(sizeof(node));
    p->children[act] = newn;
    newn->freq = 0;
    newn->ready = false;
    newn->ce = 0;
    newn->err = perr;
    newn->children = (node**)malloc(Fno*sizeof(node));; 
    for(i=0;i<=Fno-1;i++){newn->children[i] = NULL;}

    //Next child:
    if( act != last ){
        for(i=act+1;i<=last;i++){
            prob = newn->err * ((float)Item[Remap[i]] / (float)Tno );
            if ( prob >= Rminsup ){
                add_node(newn,prob,i,last);
            }
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Build the tree                                                        *
//*************************************************************************
void build_tree(){
    float prob; //probabilistic value
    int i; //loop variable
    
    //Root:
    root = (node *)malloc(sizeof(node));
    root->freq = 0;
    root->ready = false;
    root->err = 1;
    root->ce = 0;
    root->children = (node **)malloc(Fno*sizeof(node));
    for(i=0;i<=Fno-1;i++){root->children[i] = NULL;}

    //Others:
    Nodesno = 0;
    for(i=0;i<=Fno-1;i++){
        prob = (float)Item[Remap[i]] / (float)Tno;
        add_node(root,prob,i,Fno-1);
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Result to standard output                                             *
//*************************************************************************
void result_to_display(node *n, int act, int level){
    int i;
    
    if( n->freq >= Minsup ){Level[level]++;}
    for(i=act+1;i<=Fno-1;i++){
        if( n->children[i] ){
            result_to_display(n->children[i],i,level+1);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Result to display                                                     *
//*************************************************************************
void result_display(){
    int i; //loop variable

    cout << endl;
    for(i=0;i<=Fno-1;i++){Level[i] = 0;}
    result_to_display(root,-1,0);
    if( Emptytrans >= Minsup ){cout << 1 << endl;}
    else{cout << 0 << endl;}
    for(i=1;i<=Fno-1;i++){
        if( Level[i] != 0 ){cout << Level[i] << endl;}
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Result to file: frequent itemsets to the output file                          *
//*************************************************************************
void result_to_file(node *n, int act, char iname[Mask]){
    int i;
    char temp[Mask], temp2[Mask];
    
    fprintf(Fout, "%s (%d)\n", iname, n->freq);
    for(i=act+1;i<=Fno-1;i++){
        if( (n->children[i]) && (n->children[i]->freq >= Minsup) ){
            sprintf(temp2,"%d",Remap[i]);
            strcpy(temp,iname);
            strcat(temp," ");
            strcat(temp,temp2);
            result_to_file(n->children[i],i,temp);
        }
    }
}
//------------------------------------------------------------------------


//*************************************************************************
//* Result to the output file                                             *
//*************************************************************************
void result_file(){
    int i; //loop variable
    char nm[Mask]; //item name

    Fout = fopen(Foutfilename,"w");
    for(i=0;i<=Fno-1;i++){
        if( root->children[i]->freq >= Minsup ){
            sprintf(nm,"%d",Remap[i]);
            result_to_file(root->children[i],i,nm);
        }
    }
    fclose(Fout);
}
//------------------------------------------------------------------------


//*************************************************************************
//* main...                                                               *
//*************************************************************************
int main(int argc, char *argv[])
{
    int time; //for running time
                  
    //Check input parameters:
    if( argc < 3 ){
        cout << endl << "usage: fim_all datafile minsup [resultfile]" << endl;
        exit(1);
    }
    Inputfilename = argv[1]; // name of input file
    Minsup = atoi(argv[2]); // minimum support threshold
    Outputfile = false;
    if( argv[3] ){Outputfile = true; Foutfilename =argv[3];}

    if( testing ){time = clock();} //running time for testing

    read_inputdata(); //read input datafile

    //tree inic and first building phase:
    Iteration = 0; //iteration counter
    build_tree(); //build the tree
    if( testing ){cout << endl << "[" << Iteration+1 << "] " << Nodesno << " ";}
    inic_ce_freq(root,-1); //inic tree node variables
    call_count_nodefreq(); //count node->frequencies through the transactions
    set_ready(root,-1); //set node->ready variables
    if( testing ){ //just for testing
        Freqsetsno = 0;
        count_frequency(root,-1);
        cout << Freqsetsno;
        //Ready nodes:
        Readynodesno = 0;
        count_readynodes(root,-1);
        cout << " " << "(" << Readynodesno << ")";
    }

    //tree expansion phase:    
    while( !root->ready ){
        Iteration++; //increase iteration
        Alfa = Oldalfa * Alfa; //decrease alfa
        expand_tree(); //expand the tree
        if( testing ){cout << endl << "[" << Iteration+1 << "] " << Nodesno << " ";}
        inic_ce_freq(root,-1); //inic tree node variables
        call_count_nodefreq(); //count node->frequencies through the transactions
        set_ready(root,-1); //set node->ready variables
        if( testing ){ //just for testing
            Freqsetsno = 0;
            count_frequency(root,-1);
            cout << Freqsetsno;
            //Ready nodes:
            Readynodesno = 0;
            count_readynodes(root,-1);
            cout << " " << "(" << Readynodesno << ")";
        }
    }

    result_display(); //result to display
    
    if( testing ){ //running time for testing
        time = clock()-time;
	//        cout << endl << "Running time = " << (float)time / CLK_TCK << "sec" << endl;
    }
    
    if( Outputfile ){ //write the frequent itemsets to the output file (if needed):
        result_file();
        cout << endl << "Frequent sets written to " << Foutfilename << endl;
    }

}
//------------------------------------------------------------------------
